from nilmdb.utils import datetime_tz
import re

def parse_time(toparse):
    """
    Parse a free-form time string and return a datetime_tz object.
    If the string doesn't contain a timestamp, the current local
    timezone is assumed (e.g. from the TZ env var).
    """
    # If string isn't "now" and doesn't contain at least 4 digits,
    # consider it invalid.  smartparse might otherwise accept
    # empty strings and strings with just separators.
    if toparse != "now" and len(re.findall(r"\d", toparse)) < 4:
        raise ValueError("not enough digits for a timestamp")

    # Try to just parse the time as given
    try:
        return datetime_tz.datetime_tz.smartparse(toparse)
    except ValueError:
        pass

    # Try to extract a substring in a condensed format that we expect
    # to see in a filename or header comment
    res = re.search(r"(^|[^\d])("            # non-numeric or SOL
                    r"(199\d|2\d\d\d)"       # year
                    r"[-/]?"                 # separator
                    r"(0[1-9]|1[012])"       # month
                    r"[-/]?"                 # separator
                    r"([012]\d|3[01])"       # day
                    r"[-T ]?"                # separator
                    r"([01]\d|2[0-3])"       # hour
                    r"[:]?"                  # separator
                    r"([0-5]\d)"             # minute
                    r"[:]?"                  # separator
                    r"([0-5]\d)?"            # second
                    r"([-+]\d\d\d\d)?"       # timezone
                    r")", toparse)
    if res is not None:
        try:
            return datetime_tz.datetime_tz.smartparse(res.group(2))
        except ValueError:
            pass

    # Could also try to successively parse substrings, but let's
    # just give up for now.
    raise ValueError("unable to parse timestamp")

def format_time(timestamp):
    """
    Convert a Unix timestamp to a string for printing, using the
    local timezone for display (e.g. from the TZ env var).
    """
    dt = datetime_tz.datetime_tz.fromtimestamp(timestamp)
    return dt.strftime("%a, %d %b %Y %H:%M:%S.%f %z")
